/**
 *
 * DESCRIPTION
 *
 * Interrupt driven reed contact monitor with battery voltage reporting
 * for a SensebenderMicro with a reed contact from pin D3 to ground.
 *
 *******************************
 *
 * Original taken from:
 * Interrupt driven binary switch example with dual interrupts
 * Author: Patrick 'Anticimex' Fallberg
 * Connect one button or door/window reed switch between
 * digitial I/O pin 3 (REED_PIN below) and GND and the other
 * one in similar fashion on digital I/O pin 2.
 * This example is designed to fit Arduino Nano/Pro Mini
 *
 * The MySensors Arduino library handles the wireless radio link and protocol
 * between your home built sensors/actuators and HA controller of choice.
 * The sensors forms a self healing radio network with optional repeaters. Each
 * repeater and gateway builds a routing tables in EEPROM which keeps track of the
 * network topology allowing messages to be routed to nodes.
 *
 * Created by Henrik Ekblad <henrik.ekblad@mysensors.org>
 * Copyright (C) 2013-2015 Sensnology AB
 * Full contributor list: https://github.com/mysensors/Arduino/graphs/contributors
 *
 * Documentation: http://www.mysensors.org
 * Support Forum: http://forum.mysensors.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 **********************************
 * 
 * Copyright (c) 2016-2017 Michiel van der Wulp
 * 
 * 
 * V2.0 by Michiel van der Wulp
 * 
 * V2.1
 * Made ready for V2.0 MySensors libraries
 *
 * V2.2
 * Added battery voltage monitoring
 *
 * V2.3
 * Added temp & hum measurements, since the sensors are present anyhow
 * 
 * V2.4
 * First send battery, then sensors (for Domoticz bug).
 * Make LED light up at boot a bit longer.
 *
 * V2.5
 * Also send reed contact status every 30 minutes.
 * Also send the reed contact status at start-up
 * Give a description per sensor when presenting
 *
 * V2.6
 * Never ask for Acknowledge
 * 
 * V2.7
 * Reorganise interrupt, debouncing
 * 
 * V2.8
 * Do not use smartSleep - it interrupts every minute.
 * 
 * V2.9
 * Send Reed and Battery status every 6 minutes in stead of 30 minutes.
 *
 * V2.10
 * Make sure the reed contact is read every time, also when we don't expect it to be changed.
 * Do not allow interrupts during short sleep. Split the sleep call clearly between short cycle and long cycle with interrupt.
 * 
 * V2.11
 * Do not trust the interrupt result.
 * 
 * V2.12
 * Make sure that ATSHA204 is not floating.
 * Use MY_BAUD_RATE 19200 for sensebender.
 * Improve documentation.
 * 
 * V2.13
 * Send battery before every message.
 * 
 * V2.14
 * Send humidity a bit later after the temperature by outputting some debug info in between.
 * 
 * V2.15
 * Need to send temp and humidity also after a long period - since not every short period these are send - if they did not change.
 * 
 */

#define SKETCH_MAJOR_VER "2"
#define SKETCH_MINOR_VER "15"



// Enable debug prints to serial monitor
#define MY_DEBUG 
//#define MY_BAUD_RATE 115200
#define MY_BAUD_RATE 19200

// Enable and select radio type attached
#define MY_RADIO_NRF24

// The repeater does not work together with interupts
//#define MY_REPEATER_FEATURE  // Enables repeater functionality for a radio node

#include <SPI.h>
#include <MySensors.h>

#include <SI7021.h>
#include <RunningAverage.h>

#define SKETCH_NAME "Window Open Sensor"

// Child sensor ID's
#define CHILD_ID_TEMP 1
#define CHILD_ID_HUM  2
#define CHILD_ID_REED 3

// HUMI_TRANSMIT_THRESHOLD tells how much the humidity should have changed since last time it was transmitted. Likewise with
// TEMP_TRANSMIT_THRESHOLD for temperature threshold.
#define HUMI_TRANSMIT_THRESHOLD 0.5
#define TEMP_TRANSMIT_THRESHOLD 0.5
#define AVERAGES 2

#define REED_PIN 3 // Arduino Digital I/O pin for button/reed switch
// REED_PIN must be either 2 or 3 for interrupts to work

#define LED_PIN A2
#define ATSHA204_PIN   17 // A3

// How many milli seconds between each battery voltage measurement
#define WAKEUP_INTERVAL 60000 // measured in miliseconds, hence this is 1 minute
#define SEND_BATTERY 6 // hence 6 minutes


// Global settings
boolean isMetric = true;

// Storage of old measurements
long lastBattery = -100;
uint8_t lastReed = 2; // the last reported reed contact position
float lastTemperature = -100;
int lastHumidity = -100;

// Sensor messages
MyMessage msgHum(CHILD_ID_HUM, V_HUM);
MyMessage msgTemp(CHILD_ID_TEMP, V_TEMP);
MyMessage msgReed(CHILD_ID_REED, V_TRIPPED);

SI7021 humiditySensor;

RunningAverage raHum(AVERAGES);

/****************************************************
 *
 * Setup code
 *
 ****************************************************/
void setup()  
{  
  uint8_t reed;

  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, HIGH); // LED on

  // Make sure that ATSHA204 is not floating
  pinMode(ATSHA204_PIN, INPUT);
  digitalWrite(ATSHA204_PIN, HIGH);

  //Serial.begin(115200);
  Serial.print(F(SKETCH_NAME));
  Serial.print(" " SKETCH_MAJOR_VER "." SKETCH_MINOR_VER);
  Serial.flush();

  // Setup the button
  pinMode(REED_PIN, INPUT);

  // Activate internal pull-up
  digitalWrite(REED_PIN, HIGH);

  humiditySensor.begin();

  Serial.flush();
  Serial.println(F(" - Online!"));

  isMetric = getControllerConfig().isMetric;
  Serial.print(F("isMetric: ")); Serial.println(isMetric);
  raHum.clear();
  sendTempHumidityMeasurements(true);

  reed = digitalRead(REED_PIN);
  send(msgReed.set(reed==HIGH ? 1 : 0));

  sendBattLevel(true);
}

/****************************************************
 *
 * Presentation code
 *
 ****************************************************/
void presentation() {
  // Send the sketch version information to the gateway and Controller
  sendSketchInfo(SKETCH_NAME, SKETCH_MAJOR_VER "." SKETCH_MINOR_VER);

  // Register input sensors to sensor_node (they will be created as child devices)
  present(CHILD_ID_TEMP, S_TEMP, "WOS - Temperature"); // acknowledge not requested
  present(CHILD_ID_HUM,  S_HUM,  "WOS - Humidity");
  present(CHILD_ID_REED, S_DOOR, "WOS - Reed contact");
}


/***********************************************
 *
 *  Main loop function
 *
 ***********************************************/
void loop() 
{
    static int wakeupcount = -1;
    static int8_t interruptresult = -1;
    static bool shortcycleforreed = false;
    uint8_t reed;

    digitalWrite(LED_PIN, LOW); //LED off
    reed = digitalRead(REED_PIN);

    if (lastReed != reed) // the reed contact changed
    {
        // switching the LED on here allows to test the distance between the magnet and the reed contact
        digitalWrite(LED_PIN, HIGH); // LED on

        // Reed has changed from last transmission, so we send the updated reed contact value
        send(msgReed.set(reed==HIGH ? 1 : 0));

        Serial.print(F("Reed contact changed - now "));Serial.println(reed);
        Serial.flush();

        lastReed = reed;
        shortcycleforreed = true; // for debouncing, we plan to resend the reed state after a short cycle
        digitalWrite(LED_PIN, LOW); //LED off
    } 
    else
    {
        wakeupcount ++;
        if (wakeupcount > SEND_BATTERY)
        {
            Serial.println(F("Wake up - a long period has passed."));
            Serial.flush();

            sendBattLevel(true); // Not needed to send battery info that often
            reed = digitalRead(REED_PIN);
            send(msgReed.set(reed==HIGH ? 1 : 0));
            shortcycleforreed = (lastReed != reed);
            lastReed = reed;
            sendTempHumidityMeasurements(true); // V2.15: also force sending these

            wakeupcount = 0;
        } 
        else if (shortcycleforreed) 
        {
            Serial.println(F("Wake up - Short cycle after REED contact change."));
            Serial.flush();
            
            reed = digitalRead(REED_PIN);
            // Most of the time, this will send the same reed contact state as the previous time:
            send(msgReed.set(reed==HIGH ? 1 : 0));
            shortcycleforreed = (lastReed != reed);
            lastReed = reed;
        }
        else
        {
            Serial.print(F("Wake up - now minute "));Serial.println(wakeupcount);
            Serial.flush();

            sendTempHumidityMeasurements(false);
        }
    }

    if (shortcycleforreed)
    {
        // Sleep for a short while. No interrups allowed since this time acts as a debouncing period.
        interruptresult = sleep(5);
    }
    else
    {
        // Sleep until something happens with the sensor, or after a number of ms
        // Returns the interrupt number if wake up was triggered by pin change and negative if timer woke it up.
        interruptresult = sleep(REED_PIN-2, CHANGE, WAKEUP_INTERVAL);
    }
} 


/*********************************************
 *
 * Sends temperature and humidity from Si7021 sensor
 *
 * Parameters
 * - force : Forces transmission of a value (even if it's the same as previous measurement)
 *
 *********************************************/
void sendTempHumidityMeasurements(bool force)
{
  bool tx = force;

  si7021_env data = humiditySensor.getHumidityAndTemperature();

  raHum.addValue(data.humidityPercent);

  float diffTemp = abs(lastTemperature - (isMetric ? data.celsiusHundredths : data.fahrenheitHundredths)/100.0);
  float diffHum = abs(lastHumidity - raHum.getAverage());

  Serial.print(F("TempDiff :"));Serial.println(diffTemp);
  Serial.print(F("HumDiff  :"));Serial.println(diffHum);

  if (isnan(diffHum)) tx = true;
  if (diffTemp > TEMP_TRANSMIT_THRESHOLD) tx = true;
  if (diffHum > HUMI_TRANSMIT_THRESHOLD) tx = true;

  if (tx) {
    float temperature = (isMetric ? data.celsiusHundredths : data.fahrenheitHundredths) / 100.0;
    Serial.print("T: ");Serial.println(temperature);
    send(msgTemp.set(/*float value*/ temperature,/*decimals*/ 1));
    
    int humidity = data.humidityPercent;
    Serial.print("H: ");Serial.println(humidity);
    send(msgHum.set(/* int value*/ humidity));

    lastTemperature = temperature;
    lastHumidity = humidity;
  }
}


/********************************************
 *
 * Sends battery information (battery percentage)
 *
 * Parameters
 * - force : Forces transmission of a value
 *
 *******************************************/
void sendBattLevel(bool force)
{
  if (force) lastBattery = -1;

  long vcc = readVcc();
  // Calculate percentage
  vcc = vcc - 1900; // subtract 1.9V from vcc, as this is the lowest voltage we will operate at
  long percent = vcc / 14.0;
  
  if (percent != lastBattery) {
    lastBattery = percent;
    
    sendBatteryLevel(percent);
  }
}

/*******************************************
 *
 * Internal battery ADC measuring
 *
 *******************************************/
long readVcc() {
  // Read 1.1V reference against AVcc
  // set the reference to Vcc and the measurement to the internal 1.1V reference
  #if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
  ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  #elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
  ADMUX = _BV(MUX5) | _BV(MUX0);
  #elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
  ADcdMUX = _BV(MUX3) | _BV(MUX2);
  #else
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  #endif

  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA,ADSC)); // measuring

  uint8_t low  = ADCL; // must read ADCL first - it then locks ADCH
  uint8_t high = ADCH; // unlocks both

  long result = (high<<8) | low;

  result = 1125300L / result; // Calculate Vcc (in mV); 1125300 = 1.1*1023*1000
  return result; // Vcc in millivolts

}
